import os

from pydelling.config import config
from pydelling.readers.iGPReader.geometry import *
from pydelling.readers.iGPReader.io import iGPReader
from pydelling.readers.iGPReader.utils.geometry_utils import *


class PflotranExplicitWriter(iGPReader):
    """
    This class contains the functions to write a mesh file in PFLOTRAN unstructured explicit format
    """
    def write_cells(self, export_file):
        export_file.write(f"CELLS {len(self.elements)}\n")
        for element in self.elements:
            export_file.write(
                f"{element.local_id + 1} {element.centroid[0]:1.8e} {element.centroid[1]:1.8e} {element.centroid[2]:1.8e} {element.volume:1.8e}\n")

    def write_connections(self, export_file):
        # compute number of connection elements
        n_conn = 0
        for element in self.connections:
            n_conn += len(element[1])
        self.n_conn = n_conn  # set the number of connections of the mesh
        export_file.write(f"CONNECTIONS {self.n_conn}\n")
        for element in self.connections:
            prime_element = element[0]
            for connected_element in element[1]:
                face_id = element[1][connected_element]
                face_obj = self.elements[prime_element].faces[face_id]
                face_nodes = face_obj.coords
                # Compute line-face intersection
                line_points = np.array([self.elements[prime_element].centroid_coords,
                                        self.elements[connected_element].centroid_coords])
                intersection_point = line_plane_intersection(line_points=line_points,
                                                             plane_points=face_nodes)
                # compute connection centroid and area
                # intersection_point = face_obj.centroid
                conn_area = face_obj.area
                # export_file.write(f"{prime_element + 1} {connected_element + 1} {conn_centroid[0]:1.4e} {conn_centroid[1]:1.4e} {conn_centroid[2]:1.4e} {conn_area:1.4e}\n")
                export_file.write(
                    f"{prime_element + 1} {connected_element + 1} {intersection_point[0]:1.8e} {intersection_point[1]:1.8e} {intersection_point[2]:1.8e} {conn_area:1.8e}\n")

    def write_condition_data(self):
        for condition in self.region_dict:
            # Open file for dumping connection data
            if self.output_folder is None:
                file_cond = open(f"{condition}.ex", "w")
            else:
                file_cond = open(os.path.join(self.output_folder, f"{condition}.ex"), "w")
            file_cond.write(f"CONNECTIONS {self.region_dict[condition]['length']}\n")
            for id_number, element in enumerate(self.region_dict[condition]['elements']):
                conn_coords = self.nodes[element]
                if len(conn_coords == 3):
                    conn_face = TriangleFace(nodes=element,
                                             coords=conn_coords)
                elif len(conn_coords == 4):
                    conn_face = QuadrilateralFace(nodes=element,
                                                  coords=conn_coords)
                conn_area = conn_face.area
                conn_element_id = self.region_dict[condition]["centroid_id"][id_number]
                line_point = self.elements[conn_element_id].centroid_coords
                intersection_point = line_plane_perpendicular_intersection(line_point=line_point,
                                                                           plane_points=conn_coords)
                file_cond.write(
                    f"{conn_element_id + 1} {intersection_point[0]:1.8e} {intersection_point[1]:1.8e} {intersection_point[2]:1.8e} {conn_area:1.8e}\n")
            file_cond.close()

    def write_domain_postprocess_hdf5(self, export_file):

        # Create domain group
        domain_group = export_file.create_group("Domain")
        # Create Cells dataset
        _cells = []
        for element in self.elements:
            element_length = len(element.coords)
            _cells.append(config.globals.explicit_writer_dict[element_length])
            # Add elements on the _cells dataset
            list(map(lambda x: _cells.append(x), element.nodes))
        domain_group.create_dataset("Cells", data=_cells)
        # Create Vertices
        domain_group.create_dataset("Vertices", data=self.nodes)

    def write_materials(self):
        for material_name in self.material_dict:
            if self.output_folder is None:
                file_mat = open(f"{material_name}.mat", "w")
            else:
                file_mat = open(os.path.join(self.output_folder, f"{material_name}.mat"), "w")
            # Write material id elements
            list(map(lambda x: file_mat.write(f"{x}\n"), self.material_dict[material_name]))
            file_mat.close()



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pydelling.readers.iGPReader import iGPReader
import logging
from copy import deepcopy

logger = logging.getLogger(__name__)
import numpy as np
from pydelling.utils.utility_subclasses import SemistructuredFinder


class RegionOperations:
    boundaries: dict
    region_dict: dict
    # Add iGPReader class methods to this class namespace

    def divide_topography_by_z(self: iGPReader,
                               z: float,
                               region_name: str,
                               name_below='below',
                               name_above='above',
                               cache_old_results: bool = False,
                               ):
        """
        Divide the topography into two regions, one above and one below a given z coordinate.

        Args:
            z: The z coordinate to divide the topography by.
            region_name: The name of the region to divide.
            name_below: The name of the region below the z coordinate.
            name_above: The name of the region above the z coordinate.
            cache_old_results: If True, each time this method is called the original region_dict will be used.

        Returns:
            None
        """
        if cache_old_results:
            if hasattr(self, 'old_region_dict'):
                self.region_dict = self.old_region_dict
            if hasattr(self, 'old_boundaries'):
                self.boundaries = self.old_boundaries
        boundary_faces = self.get_boundary_faces(region_name)
        name_below = name_below if name_below != 'below' else f"{region_name}_below"
        name_above = name_above if name_above != 'above' else f"{region_name}_above"
        logger.info(f'Dividing topography by z={z} into regions {name_below} and {name_above}')

        new_boundaries = {name_below: [],
                          name_above: []}

        for face_idx, face in enumerate(boundary_faces):
            if face.centroid[2] < z:
                new_boundaries[name_below].append(face_idx)
            else:
                new_boundaries[name_above].append(face_idx)

        # Update the boundaries dict
        if cache_old_results:
            self.old_boundaries = deepcopy(self.boundaries)
        for name in new_boundaries:
            self.boundaries[name] = [boundary_faces[idx] for idx in new_boundaries[name]]
        self.boundaries.pop(region_name)

        # Update the region dict
        if cache_old_results:
            self.old_region_dict = deepcopy(self.region_dict)
        get_old_region = self.region_dict[region_name]
        # Divide the region elements based on the found idx
        self.region_dict[name_below] = {'elements': get_old_region['elements'][new_boundaries[name_below]],
                                        'centroid_id': get_old_region['centroid_id'][new_boundaries[name_below] ],
                                        'length': len(new_boundaries[name_below])}
        self.region_dict[name_above] = {'elements': get_old_region['elements'][new_boundaries[name_above]],
                                        'centroid_id': get_old_region['centroid_id'][new_boundaries[name_above]],
                                        'length': len(new_boundaries[name_above])}

        # Remove the old region
        self.region_dict.pop(region_name)

    def get_nodes_from_x_y(self: iGPReader,
                           x: float,
                           y: float,
                           materials: list = None,
                           top_region_name: str = None,
                           tol: float = 5.0,
                           eps: float = 2.5,
                           min_samples: int = 20,
                           ) -> list:
        """
        Returns a list of nodes closest to the specified x and y coordinates.

        Args:
            self: The iGPReader object that calls this method.
            x: The x-coordinate for which to find closest nodes.
            y: The y-coordinate for which to find closest nodes.
            materials: The materials of the nodes to consider.
            top_region_name: The top region name of the nodes to consider.
            tol: The tolerance used in the computation.
            eps: The epsilon value used in clustering.
            min_samples: The minimum number of samples required for a cluster.

        Returns:
            A list of node IDs closest to the specified x and y coordinates.
        """


        if not hasattr(self, 'is_subset_generated'):
            self._setup_cluster_and_subset(eps=eps,
                                           min_samples=min_samples,
                                           top_region_name=top_region_name,
                                           materials=materials,
                                           )
            self.is_subset_generated = True

        closest_cluster_ids = self.cluster_engine.get_closest_point_ids_from_xy(x, y)
        closest_cluster_nodes = self.nodes[self.node_id_subset[closest_cluster_ids]]
        # Sort the closest_cluster_ids by z coordinate


        return self.node_id_subset[closest_cluster_ids]
        # Order


    def _setup_cluster_and_subset(self: iGPReader,
                                    eps: float,
                                    min_samples: int,
                                    top_region_name: str = None,
                                    materials: list = None,
                                    ):

        """
        Initialize the cluster engine (DBSCAN) and create a subset of nodes from the material specified

        Args:
            eps: The maximum distance between two samples for one to be considered as in the neighborhood of the other
            min_samples: The number of samples in a neighborhood for a point to be considered as a core point.
            top_region_name: The name of the top region.
            materials: List of materials.

        """
        logger.info(f'Setting up cluster engine (DBSCAN with eps={eps} and min_samples={min_samples}) and subset of nodes for materials {materials}')
        self.is_subset_generated = False
        temp_node_id = []
        if materials is not None:
            for material in materials:
                element_nodes = [self.element_nodes[element_id] for element_id in self.get_material_elements(material)]
                element_nodes = np.unique(np.array(element_nodes).flatten())
                temp_node_id.append(element_nodes)
        else:
            element_nodes = np.unique(np.array(self.element_nodes).flatten())
            temp_node_id.append(element_nodes)
        temp_node_id = np.concatenate(temp_node_id)
        self.node_id_subset = np.unique(temp_node_id)
        self.cluster_engine = SemistructuredFinder(self.nodes[self.node_id_subset], eps=eps, min_samples=min_samples)
        logger.info(f'A total of {self.cluster_engine.n_clusters} clusters were found in the subset of nodes. Play with eps (eps={eps}) and min_samples (min_samples={min_samples}) to find more clusters.')
        try:
            top_region_elements = self.get_region_nodes(top_region_name)
            if self.cluster_engine.n_clusters != len(top_region_elements):
                logger.warning(f'The number of clusters found ({self.cluster_engine.n_clusters}) does not match the number of elements in the top region ({len(top_region_elements)}).')
        except:
            logger.warning(f'No top region was provided. The number of clusters found ({self.cluster_engine.n_clusters}) may not match the number of elements in the top region.')




import os
import shutil
from pathlib import Path
from typing import Dict

import h5py
import numpy as np


class PostProcessingEngine:
    """This class manages the structure of the output files"""
    def __init__(self, dt=0.2, unit="y"):
        self.dt = dt
        self.unit = unit
        self.output_times = []
        self.output_units = []

    def run(self):
        "Method that runs the post-processing engine"
        self.find_directories()
        self.find_output_files()
        self.find_domain_file()
        self.find_input_file()
        if len(self.output_vtk_vel_files) != 0:
            self.add_velocity_to_hdf5()
        self.find_attributes()
        self.export_xmf()
        self.copy_domain_file()

    def find_output_files(self):
        """This method searches the current directory and the directory called "./output-hdf5" to find PFLOTRAN output files"""
        # First, see if the directory contains the output files
        if self.current_directory.stem == "output-hdf5":
            self.output_h5_files = list(self.current_directory.glob("*.h5"))
            self.output_directory = self.current_directory
            assert len(self.output_h5_files) > 0, "Output hdf5 files couldn't be found"
        else:
            self.output_h5_files = list(self.root_directory.glob("*.h5"))
            self.output_directory = self.current_directory
            if len(self.output_h5_files) == 0:
                # Check the output-hdf5 folder
                output_folder = self.current_directory / "output-hdf5"
                self.output_h5_files = list(output_folder.glob("*.h5"))
                # assert len(self.output_h5_files) > 0, "Output hdf5 files couldn't be found"
                self.output_directory = self.current_directory / "output-hdf5"
        # Take out the domain hdf5 file from the output file list
        domain_file = [file for file in self.output_h5_files if "domain" in file.name]
        if domain_file:
            self.output_h5_files.remove(domain_file[0])
        self.n_output_files = len(self.output_h5_files)
        self.output_h5_files = sorted(self.output_h5_files)
        # Process the VTK files
        if self.current_directory.stem == "output-vtk":
            self.output_vtk_vel_files = list(self.current_directory.glob("*vel*.vtk"))
            self.output_vtk_directory = self.current_directory
            assert len(self.output_vtk_vel_files) > 0, "Output hdf5 files couldn't be found"
        else:
            self.output_vtk_vel_files = list(self.root_directory.glob("*vel*.vtk"))
            self.output_vtk_directory = self.current_directory
            if len(self.output_vtk_vel_files) == 0:
                # Check the output-hdf5 folder
                output_folder = self.current_directory / "output-vtk"
                self.output_vtk_vel_files = list(output_folder.glob("*vel*.vtk"))
                # assert len(self.output_vtk_files) > 0, "Output vtk files couldn't be found"
                self.output_vtk_directory = self.current_directory / "output-vtk"
        self.output_vtk_vel_files = sorted(self.output_vtk_vel_files)

    def find_directories(self):
        """Test """
        self.current_directory = Path.cwd()
        if self.current_directory.stem == "output-hdf5":
            self.root_directory = self.current_directory.parent
        elif self.current_directory.glob("*.in"):
            self.root_directory = self.current_directory
        elif self.current_directory.stem == "input_files":
            self.root_directory = self.current_directory.parent
        else:
            raise FileNotFoundError("Root directory couldn't be found")
        self.input_directory = self.root_directory / "input_files"

    def find_domain_file(self):
        self.domain_file = list(self.input_directory.glob("*-domain.h5"))[0]
        return self.domain_file

    def find_input_file(self):
        self.input_file = list(self.root_directory.glob("*.in"))[0]
        return self.input_file

    def add_velocity_to_hdf5(self):
        for i in range(0, self.n_output_files):
            with h5py.File(self.output_h5_files[i], "r+") as h5_file:
                vtk_filename = self.output_vtk_vel_files[i]
                vtk_file_dict = self.read_vtk_file(filename=vtk_filename)
                for var in vtk_file_dict:
                    if var == "Material_ID":
                        continue
                    hdf5_group: h5py.Group = list(h5_file.values())[0]
                    var_names = [name for name in hdf5_group]
                    if not var in var_names:
                        hdf5_group.create_dataset(name=var, data=vtk_file_dict[var]["data"])

    def find_attributes(self):
        self.input_stem = self.input_file.stem
        # Find number of cells and number of vertices
        with h5py.File(self.domain_file, "r") as file:
            self.n_vertices = len(file["Domain"]["Vertices"])
        # Find output variables
        with h5py.File(self.output_h5_files[0], "r") as hdf5_file:
            hdf5_group: h5py.Group = list(hdf5_file.values())[0]
            self.output_variables = [item[0] for item in hdf5_group.items()]
            self.n_cells = len(hdf5_group[self.output_variables[0]])
        for hdf5_file in self.output_h5_files:
            with h5py.File(hdf5_file, "r") as hdf5_file:
                attribute_list = list(hdf5_file.items())[0][0].split()
                self.output_times.append(float(attribute_list[2]))
                self.output_units.append(attribute_list[3])

    def copy_domain_file(self):
        # check if file exists and delete otherwise
        if list(self.output_directory.glob("*-domain.h5")):
            os.remove(list(self.output_directory.glob("*-domain.h5"))[0])
        shutil.copy2(self.domain_file, self.output_directory / f"{self.input_stem}-domain.h5")

    def export_xmf(self):
        # Input variables
        dt = self.dt  # time interval between individual result files
        file_name = self.input_stem
        Nelements = self.n_cells
        Nvertices = 6 * self.n_cells
        Nvertices_pflotran = self.n_vertices
        print(Nelements, Nvertices, Nvertices_pflotran)

        # Script to generate the files
        for i in range(0, self.n_output_files):
            self.name = f"{file_name}-{i:03d}.xmf"
            self.time = dt * i
            name_h5 = f"{file_name}-{i:03d}.h5"
            self.export_file = self.output_directory
            with open(self.output_directory / self.name, 'w') as self.output_file:
                self.output_file.write('<?xml version="1.0" ?>\n')
                self.output_file.write('<!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>\n')
                self.output_file.write('<Xdmf>\n')
                self.output_file.write('\t<Domain>\n')
                self.output_file.write('\t<Grid Name="Mesh">\n')
                self.output_file.write(f'\t\t<Time Value = "{self.output_times[i]:1.5E}" />\n')
                self.output_file.write('\t\t<Topology Type="Mixed" NumberOfElements="%s" >\n' % Nelements)
                self.output_file.write('\t\t\t<DataItem Format="HDF" DataType="Int" Dimensions="%s">\n' % (Nelements + Nvertices))
                self.output_file.write('\t\t\t\t %s-domain.h5:/Domain/Cells\n' % file_name)
                self.output_file.write('\t\t\t</DataItem>\n')
                self.output_file.write('\t\t</Topology>\n')
                self.output_file.write('\t\t<Geometry GeometryType="XYZ">\n')
                self.output_file.write('\t\t\t<DataItem Format="HDF" Dimensions="%s 3">\n' % Nvertices_pflotran)
                self.output_file.write('\t\t\t\t %s-domain.h5:/Domain/Vertices\n' % file_name)
                self.output_file.write('\t\t\t</DataItem>\n')
                self.output_file.write('\t\t</Geometry>\n')
                for output_variable in self.output_variables:
                    self.export_attribute(output_variable, i)
                self.output_file.write('\t</Grid>\n')
                self.output_file.write('\t</Domain>\n')
                self.output_file.write('</Xdmf>\n')

    def export_attribute(self, var, i):
        self.output_file.write('\t\t<Attribute Name="%s" AttributeType="Scalar"  Center="Cell">\n' % var)
        self.output_file.write('\t\t\t<DataItem Dimensions="%s 1" Format="HDF">\n' % self.n_cells)
        self.output_file.write(f'\t\t\t\t{self.input_stem}-{i:03d}.h5:/   {i} Time  {self.output_times[i]:1.5E} {self.output_units[i]}/{var}\n')
        self.output_file.write('\t\t\t</DataItem>\n')
        self.output_file.write('\t\t</Attribute>\n')

    @staticmethod
    def read_vtk_file(filename) -> Dict:
        """This method reads an VTK file and returns the data"""
        # Process vtk file
        print(f"Processing {filename} VTK file")
        data_dict = {}
        temp_array = []
        start_reading = False
        # filename = "G:/My Drive/dev/2020-project_repos/rwm_repo/scripts/3-post_processing/B-VTK_velocity/output-vtk/test-trial.vtk"
        with open(filename, "r") as vtk_file:
            for line in vtk_file.readlines():
                split_line = line.split()
                if not split_line:
                    start_reading = False
                if "CELL_DATA" in split_line:
                    n_cells = split_line[1]
                if "SCALARS" in split_line:
                    data_dict[split_line[1]] = {"name": split_line[1],
                                                "n_cells": n_cells,
                                                "data": []
                                                }
                    try:
                        data_dict[current_reading]["data"] = np.concatenate(temp_array)
                        data_dict[current_reading]["shape"] = data_dict[current_reading]["data"].shape
                    except:
                        pass
                    start_reading = False
                    current_reading = split_line[1]
                    print(f"Starting to read variable {current_reading}")
                if "LOOKUP_TABLE" in split_line:
                    start_reading = True
                    temp_array = []
                    continue
                if start_reading:
                    temp_array.append(np.array(split_line).astype(np.float))
            data_dict[current_reading]["data"] = np.concatenate(temp_array)
            data_dict[current_reading]["shape"] = np.concatenate(temp_array).shape
        return data_dict


from .BaseStudy import BaseStudy
import logging
from typing import Union, List, Dict

logger = logging.getLogger(__name__)


class PflotranStudy(BaseStudy):
    """This class extends the BaseStudy class to manage PFLOTRAN related simulations.
    """
    def __init__(self, input_file: str, *args, **kwargs):
        """This method initializes the class.
        """
        super().__init__(input_file, *args, **kwargs)
        self.regions_to_idx = {}
        self.datasets_to_idx = {}

    def get_regions(self):
        """This method returns the regions of the simulation.
        """
        region_lines = self._find_tags('region')
        regions = []
        for line_idx in region_lines:
            line = self._get_line(line_idx)
            if self._get_parent_tag_name_(line_idx).lower() == 'region':
                regions.append(line.split()[1])
                self.regions_to_idx[line.split()[1]] = line_idx
        return regions

    def get_simulation_time(self, time_unit: str = 'y'):
        """This method returns the output time series of the simulation"""
        time_lines = self._get_line(self._find_tags('FINAL_TIME')[0])
        value = time_lines.split()[1]
        value_unit = time_lines.split()[2]
        return self.convert_time(value=value, initial_unit=value_unit, final_unit=time_unit)

    def get_checkpoint(self):
        """This method returns the checkpoint times of the simulation.
        """
        checkpoint_lines = self._find_tags('CHECKPOINT')
        if len(checkpoint_lines) == 0:
            return None
        checkpoint_line = checkpoint_lines[0]
        checkpoint_block = self._get_block_lines(checkpoint_line)
        for line in checkpoint_block:
            if 'times' in line.lower():
                return line
        return None

    def replace_simulation_time(self, new_time: float, time_unit: str = 'y'):
        """This method replaces the simulation time of the simulation.
        """
        time_lines = self._find_tags('FINAL_TIME')
        new_time = self.convert_time(value=new_time, initial_unit=time_unit, final_unit=time_unit)
        self._replace_line(line_index=time_lines[0], new_line=['FINAL_TIME', str(new_time), time_unit])

    def get_region_file(self, region: str) -> Union[str, None]:
        """This method returns the file of the region.
        """
        self.get_regions()
        region_line = self.regions_to_idx[region]
        region_block = self._get_block_lines(region_line)
        for line in region_block:
            if 'file' in line.lower():
                return line.split()[1]
        return None

    def get_datasets(self) -> List[str]:
        """This method returns the datasets of the simulation.
        """
        datasets = []
        for line_idx in self._find_tags('DATASET'):
            line = self._get_line(line_idx)
            if line.split()[0].lower() == 'hdf5_dataset_name':
                continue
            if self._get_parent_tag_name_(line_idx).lower() == 'dataset':
                datasets.append(line.split()[1])
                self.datasets_to_idx[line.split()[1]] = line_idx
        return datasets



    def replace_region_file(self, region: str, new_file: str):
        """This method replaces the file of the region.
        """
        self.get_regions()
        old_file = self.get_region_file(region)
        region_line = self.regions_to_idx[region]
        region_block = self._get_block_lines(region_line)
        region_block_idx = self._get_block_line_idx(region_line)
        logger.info(f"Replacing region {region} file from '{old_file}' to '{new_file}'")
        for line_idx, line in zip(region_block_idx, region_block):
            if 'file' in line.lower():
                self._replace_line(line_index=line_idx, new_line=['FILE', new_file])

    def add_checkpoint(self, times: Union[float, List[float]], time_unit: str = 'y'):
        """This method adds a checkpoint to the simulation.
        """
        logger.info(f"Adding checkpoint at {times} {time_unit}")
        if isinstance(times, float):
            times = [times]
        times = [self.convert_time(value=time, initial_unit=time_unit, final_unit=time_unit) for time in times]
        times = [str(time) for time in times]
        # Check if the checkpoint tag is already in the input file
        if self.has_tag('CHECKPOINT'):
            checkpoint_line = self._find_tags('CHECKPOINT')[0]
            checkpoint_block = self._get_block_lines(checkpoint_line)
            checkpoint_block_idx = self._get_block_line_idx(checkpoint_line)
            for line_idx, line in zip(checkpoint_block_idx, checkpoint_block):
                if 'times' in line.lower():
                    self._replace_line(line_index=line_idx, new_line=['TIMES', time_unit, *times])
        else:
            # Find simulation block
            simulation_block_idx = self._get_block_line_idx(self._find_tags('SIMULATION')[0])
            # Find the last line of the simulation block
            last_line_idx = simulation_block_idx[-1] - 1
            # Add the checkpoint block
            self._add_line(line_index=last_line_idx, new_line=['CHECKPOINT'])
            self._add_line(line_index=last_line_idx + 1, new_line=['TIMES', time_unit, *times])
            self._add_line(line_index=last_line_idx + 2, new_line=['FORMAT', 'HDF5'])
            self._add_line(line_index=last_line_idx + 3, new_line=['/'])

    def add_restart(self, filename: str):
        # Find simulation block
        simulation_block_idx = self._get_block_line_idx(self._find_tags('SIMULATION')[0])
        # Find the last line of the simulation block
        last_line_idx = simulation_block_idx[-1]
        # Add the checkpoint block
        self._add_line(line_index=last_line_idx, new_line=['RESTART'])
        self._add_line(line_index=last_line_idx + 1, new_line=['FILENAME', filename])
        self._add_line(line_index=last_line_idx + 2, new_line=['/'])

    def add_dataset(self, name: str, filename: str, hdf5_dataset_name: str):
        """This method adds a dataset to the simulation.
        """
        logger.info(f"Adding dataset {name} to the simulation")
        # Find simulation block
        if name in self.get_datasets():
            idx = self.datasets_to_idx[name]
            inside_block = self._get_block_line_idx(idx)
            for line_idx in inside_block:
                line = self._get_line(line_idx)
                if 'file' in line.lower():
                    self._replace_line(line_index=line_idx, new_line=['FILENAME', filename])
                if 'hdf5_dataset' in line.lower():
                    self._replace_line(line_index=line_idx, new_line=['HDF5_DATASET_NAME', hdf5_dataset_name])
        else:
            logger.warning(f"Dataset {name} not found in the simulation. Adding it.")
            subsurface_block_start = self.get_subsurface_idx()
            # Find the last line of the simulation block
            last_line_idx = subsurface_block_start + 2
            # Add the checkpoint block
            self._add_line(line_index=last_line_idx, new_line=['# Automatically added by pydelling'])
            self._add_line(line_index=last_line_idx + 1, new_line=['DATASET', name])
            self._add_line(line_index=last_line_idx + 2, new_line=['FILENAME', filename])
            self._add_line(line_index=last_line_idx + 3, new_line=['HDF5_DATASET_NAME', hdf5_dataset_name])
            self._add_line(line_index=last_line_idx + 4, new_line=['END'])

    def get_subsurface_idx(self) -> int:
        """This method returns the index of the subsurface tag.
        """
        subsurface_idx = self._find_tags('SUBSURFACE')
        for idx in subsurface_idx:
            if self._get_parent_tag_name_(idx) == 'SUBSURFACE':
                return idx

    def has_tag(self, tag: str):
        """This method returns True if the tag is in the input file.
        """
        return len(self._find_tags(tag)) > 0

    def _get_parent_tag_name_(self, line_index: int):
        """This method returns the parent tag of the line.
        """
        # Find the previous END tag
        has_end_tag = False
        temp_list = []
        temp_list.append(self._get_line(line_index))
        while not has_end_tag:
            line = self._get_line(line_index)
            line_index -= 1
            if len(line.split()) == 0:
                continue
            if line[0] == '#':
                continue
            if 'subsurface' in line.lower():
                has_end_tag = True
            temp_list.append(line)
            if 'end' in line.lower():
                has_end_tag = True

        return temp_list[-2].split()[0]

    def _get_block_lines(self, line_index: int):
        """This method returns the lines of the block.
        """
        # Find the previous END tag
        has_end_tag = False
        temp_list = []
        temp_list.append(self._get_line(line_index))
        while not has_end_tag:
            line_index += 1
            line = self._get_line(line_index)
            if len(line.split()) == 0:
                continue
            if line[0] == '#':
                continue
            temp_list.append(line)
            if 'end' in line.lower():
                has_end_tag = True
        return temp_list

    def _get_block_line_idx(self, line_index: int):
        """This method returns the lines of the block.
        """
        # Find the previous END tag
        has_end_tag = False
        temp_list = []
        temp_list.append(line_index)
        while not has_end_tag:
            line_index += 1
            line = self._get_line(line_index)
            if len(line.split()) == 0:
                continue
            if line[0] == '#':
                continue
            temp_list.append(line_index)
            if 'end' in line.lower():
                has_end_tag = True
        return temp_list

    # Class properties
    @property
    def final_time(self):
        """This method returns the final time of the simulation.
        """
        return self.get_simulation_time()

    @property
    def final_time_unit(self):
        """This method returns the final time unit of the simulation.
        """
        return self._get_line(self._find_tags('FINAL_TIME')[0]).split()[2]

    @property
    def idx_to_regions(self):
        """This method returns the regions of the simulation.
        """
        return {v: k for k, v in self.regions_to_idx.items()}






import unittest
from pydelling.utils.utility_subclasses import SemistructuredFinder
import numpy as np

class GeometryUtilsCase(unittest.TestCase):
    def test_semistructured_finder_2d(self):
        n_clusters_x = 10
        n_clusters_y = 10
        n_points = 25
        cluster_centers = [[x, y] for x in range(n_clusters_x) for y in range(n_clusters_y)]
        points = [[x + np.random.normal(0, 0.02), y + np.random.normal(0, 0.02)] for x, y in cluster_centers for _ in range(n_points)]
        points = np.array(points)

        cluster_finder = SemistructuredFinder(points, eps=0.05, min_samples=5)
        self.assertEqual(cluster_finder.n_clusters, n_clusters_x * n_clusters_y)

    def test_semistructured_finder_3d(self):
        n_clusters_x = 10
        n_clusters_y = 10
        n_points = 25
        cluster_centers = [[x, y, z] for x in range(n_clusters_x) for y in range(n_clusters_y) for z in range(10)]
        points = [[x + np.random.normal(0, 0.02), y + np.random.normal(0, 0.02), z] for x, y, z in cluster_centers for _ in
                  range(n_points)]
        points = np.array(points)

        cluster_finder = SemistructuredFinder(points, eps=0.05, min_samples=20, projection_axis='z')
        self.assertEqual(cluster_finder.n_clusters, n_clusters_x * n_clusters_y)

        # # Plot the clusters
        # import matplotlib.pyplot as plt
        # for cluster in cluster_finder.clusters:
        #     # Transform ((x1, y1), (x2, y2), ...) to (x1, x2, ...), (y1, y2, ...)
        #     x, y, z = zip(*cluster)
        #     plt.scatter(x, y)
        # plt.show()

        # # Plot just the first cluster
        # first_cluster = cluster_finder.clusters[5]
        # x, y, z = zip(*first_cluster)
        # import matplotlib.pyplot as plt
        # plt.scatter(x, y)
        # plt.show()


    def test_semistructured_finder_point_columns(self):
        n_clusters_x = 10
        n_clusters_y = 10
        n_points = 25
        cluster_centers = [[x, y, z] for x in range(n_clusters_x) for y in range(n_clusters_y) for z in range(10)]
        points = [[x + np.random.normal(0, 0.02), y + np.random.normal(0, 0.02), z] for x, y, z in cluster_centers for _ in
                  range(n_points)]
        points = np.array(points)

        cluster_finder = SemistructuredFinder(points, eps=0.05, min_samples=20, projection_axis='z')
        target_point = np.array([5.0, 5, 5])
        closest_cluster = cluster_finder.get_closest_cluster_from_point(target_point)
        self.assertLess(np.linalg.norm(target_point - np.array(closest_cluster).mean()), 0.3)
        # Plot the clusters with high alpha, the selected cluster opaque and the point
        # import matplotlib.pyplot as plt
        # x, y, z = zip(*closest_cluster)
        # plt.scatter(x, y)
        # plt.scatter(target_point[0], target_point[1], alpha=1)
        # plt.show()

        closest_cluster = cluster_finder.get_closest_cluster_from_xy(x=5.0, y=5.0)
        self.assertLess(np.linalg.norm(target_point - np.array(closest_cluster).mean()), 0.3)

        # Get the point ids instead of the points
        closest_cluster = cluster_finder.get_closest_point_ids_from_point(target_point)









import streamlit as st
import extra_streamlit_components as stx

class BaseStreamlitUtilityClass:
    def save_in_session_state(self, key, value: object):
        """This method saves the object to a session state"""
        st.session_state[key] = value

    def initialize_in_session_state(self, key: str, value: object=None):
        """This method initializes the session state"""
        if key not in st.session_state:
            st.session_state[key] = value

    def get_from_session_state(self, key: str):
        """This method gets the session state"""
        return st.session_state[key]

    def initialize(self):
        """This method initializes the webapp."""
        pass

    def set_to_session_state(self, key: str, value: object):
        """This method equals the previous method"""
        self.save_in_session_state(key, value)

    @staticmethod
    @st.cache(allow_output_mutation=True, suppress_st_warning=True)
    def get_manager():
        return stx.CookieManager()


from .BaseStatus import BaseStatus
import logging
import re

logger = logging.getLogger(__name__)

class PflotranStatus(BaseStatus):
    """This class reads a pflotran status file and extracts some key information."""
    def __init__(self,
                 status_file,
                 total_time=None):
        super().__init__(status_file)
        self.times = []
        self.dts = []
        self.wall_clock_time = None
        self.total_time = total_time
        self.extract_key_information()

    def extract_key_information(self):
        """Extract total Time, Dt, and Wall Clock Time values from the status data."""

        # Search for total Time and Dt values
        pattern = re.compile(r"== REACTIVE TRANSPORT =+.*?Step.*?Time=\s+([\d\.E\+\-]+).*?Dt=\s+([\d\.E\+\-]+)",
                             re.DOTALL)
        matches = pattern.findall(self.status_data)
        for match in matches:
            self.times.append(float(match[0]))
            self.dts.append(float(match[1]))

        # Search for Wall Clock Time value
        pattern = re.compile(r"Wall Clock Time:\s+([\d\.E\+\-]+)")
        match = pattern.search(self.status_data)
        if match:
            self.wall_clock_time = float(match.group(1))

    @property
    def progress(self):
        """Return the progress of the simulation."""
        if self.total_time is None:
            return None
        return self.times[-1] / self.total_time

    @property
    def is_done(self):
        """Return whether the simulation is done."""
        if self.wall_clock_time is None:
            return False
        else:
            return True

    def add_total_time(self, total_time):
        """Add the total time of the simulation."""
        self.total_time = total_time

    def read(self, status_file = None):
        """Read the status file and extract key information."""
        self.status_file = status_file if status_file is not None else self.status_file
        self.read_status_file()
        self.extract_key_information()



